#include <complex>
#include <cmath>

#include <Eigen/Dense>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>

using namespace std;
using namespace Eigen;
using namespace HARK;

typedef float Real;
//typedef complex<Real> Complex;

typedef Eigen::Matrix<Real, Dynamic, Dynamic> MatrixXr;
//typedef Eigen::Matrix<Complex, Dynamic, Dynamic> MatrixXc;
typedef Eigen::Matrix<complex<Real>, Dynamic, Dynamic> MatrixXc;

typedef Eigen::Matrix<Real, Dynamic, 1> VectorXr;
//typedef Eigen::Matrix<Complex, Dynamic, 1> VectorXc;
typedef Eigen::Matrix<complex<Real>, Dynamic, 1> VectorXc;

#include <rpca.hh>

class MultiRobustPCA;

DECLARE_NODE(MultiRobustPCA);
/*Node
 *
 * @name MultiRobustPCA
 * @category HARK:Rescue
 * @description Robust PCA
 *
 * @input_name INPUT
 * @input_type Matrix<complex<float> >
 * @input_description
 *
 * @output_name OUTPUT
 * @output_type Matrix<float>
 * @output_description
 *
 * @parameter_name NUM_CHANNEL
 * @parameter_type int
 * @parameter_value 8
 * @parameter_description 
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description 
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description
 *
 * @parameter_name CUTOFF_FREQ
 * @parameter_type float
 * @parameter_value 125
 * @parameter_description 
 *
 * @parameter_name MAX_LANK
 * @parameter_type int
 * @parameter_value 100
 * @parameter_description
 *
 * @parameter_name GLOBAL_SCALE
 * @parameter_type float
 * @parameter_value 5e-2
 * @parameter_description 
 *
 * @parameter_name SCALE_UPDATE_RATIO
 * @parameter_type float
 * @parameter_value 0.01
 * @parameter_description 
 *
 * @parameter_name LAMBDA
 * @parameter_type float
 * @parameter_value 1e-1
 * @parameter_description 
 *
 * @parameter_name EPSILON
 * @parameter_type float
 * @parameter_value 1e-3
 * @parameter_description 
 *
END*/

class MultiRobustPCA : public BufferedNode 
{
  int id_input;
  int id_output;

  int M, F, K;

  int cutoff_bin;
  Real scale;
  Real sur;

  vector<RobustPCA> rpca;

  MatrixXr scale2;

  MatrixXr amp_spec;

public:
  MultiRobustPCA(string nodeName, ParameterSet params) :
    BufferedNode(nodeName, params)
  {
    id_input  = addInput("INPUT");
    id_output = addOutput("OUTPUT");

    M = dereference_cast<int>(parameters.get("NUM_CHANNEL"));

    F = dereference_cast<int>(parameters.get("LENGTH")) / 2 + 1;
    cutoff_bin = F * dereference_cast<float>(parameters.get("CUTOFF_FREQ")) 
      / ((float)dereference_cast<int>(parameters.get("SAMPLING_RATE")) / 2);

    scale = dereference_cast<float>(parameters.get("GLOBAL_SCALE"));

    Real lm = dereference_cast<float>(parameters.get("LAMBDA"));

    sur = dereference_cast<float>(parameters.get("SCALE_UPDATE_RATIO"));

    scale2 = MatrixXr::Zero(F, M);

    rpca.assign(M, RobustPCA(F, dereference_cast<int>(parameters.get("MAX_LANK"))));
    float ep = dereference_cast<float>(parameters.get("EPSILON"));
    for (int m=0;m<M;m++) {
      rpca[m].initialize();
      rpca[m].lm1 = rpca[m].lm2 = 1 / sqrt(F) * lm;
      rpca[m].ep = ep;
    }

    amp_spec.setZero(F, M);

    initParallel();
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    // translate RCPtr to Eigen
//    RCPtr<HARK::Matrix<Complex> > p_input = getInput(id_input, count);
    RCPtr<HARK::Matrix<complex<Real>> > p_input = getInput(id_input, count);
    Eigen::Map<const MatrixXc> input((*p_input)[0], F, M);

    RCPtr<HARK::Matrix<Real> > p_output(new HARK::Matrix<Real>(M, F));
    Eigen::Map<MatrixXr> output((*p_output)[0], F, M);

    (*(outputs[id_output].buffer))[count] = p_output;

    // conduct RPCA at each channel
#pragma omp parallel for
    for (int m=0;m<M;m++) {
      amp_spec.col(m) = input.col(m).cwiseAbs();

      if (count == 0) {
	scale2.col(m) = amp_spec.col(m);
      } else {
	scale2.col(m) = (1 - sur) * scale2.col(m) + sur * amp_spec.col(m);
      }

      amp_spec.col(m) = scale * amp_spec.col(m).cwiseQuotient(scale2.col(m));

      amp_spec.block(0, m, cutoff_bin, 1) = MatrixXr::Constant(cutoff_bin, 1, 0);
      
      rpca[m].update(amp_spec.col(m));

      output.col(m) = 1 / scale * rpca[m].e
	.unaryExpr([](float v)->float{ return v > 0 ? v : 0;})
	.cwiseProduct(scale2.col(m));
    }
  }
};
