#include <complex>
#include <cmath>

#include <Eigen/Dense>
//#include <boost/math/special_functions/digamma.hpp>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>

using namespace std;
using namespace Eigen;
using namespace HARK;

typedef float Real;
//typedef complex<Real> Complex;

typedef Eigen::Matrix<Real, Dynamic, Dynamic> MatrixXr;
//typedef Eigen::Matrix<Complex, Dynamic, Dynamic> MatrixXc;
typedef Eigen::Matrix<complex<Real>, Dynamic, Dynamic> MatrixXc;

typedef Eigen::Matrix<Real, Dynamic, 1> VectorXr;
//typedef Eigen::Matrix<Complex, Dynamic, 1> VectorXc;
typedef Eigen::Matrix<complex<Real>, Dynamic, 1> VectorXc;

class PoissonGammaIntegrator;

DECLARE_NODE(PoissonGammaIntegrator);
/*Node
 *
 * @name PoissonGammaIntegrator
 * @category HARK:Rescue
 * @description Robust PCA
 *
 * @input_name SPEC
 * @input_type Matrix<complex<float> >
 * @input_description
 *
 * @input_name SPARSE
 * @input_type Matrix<float>
 * @input_description
 *
 * @output_name OUTPUT
 * @output_type Matrix<complex<float> >
 * @output_description
 *
 * @parameter_name NUM_CHANNEL
 * @parameter_type int
 * @parameter_value 8
 * @parameter_description  
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description
 *
 * @parameter_name GLOBAL_SCALE
 * @parameter_type float
 * @parameter_value 100
 * @parameter_description 
 *
END*/

class PoissonGammaIntegrator : public BufferedNode
{
  Real EPS = 1e-12;

  int id_input_spec;
  int id_input_sparse;
  int id_output;

  int M, F;

  Real scale;

  Real a0;
  VectorXr b0;
  Real a1, b1;
  Real a2, b2;

  VectorXr s_a, s_b;

  VectorXr g_a;
  Real g_b;
  
  MatrixXr e_a;
  Real e_b;

  static float digamma(float x){
    float x2 = 2.0f + x;

    union {
      float f;
      uint32_t i;
    } vx = { x2 };

    union {
      uint32_t i;
      float f;
    } mx = { (vx.i & 0x007FFFFF) | 0x3f000000 };

    float y = vx.i;
    y *= 1.1920928955078125e-7f;

    float logterm = 0.69314718 *
      (y - 124.22551499f
         - 1.498030302f * mx.f
         - 1.72587999f / (0.3520887068f + mx.f));

    return (-48.0f + x * (-157.0f + x * (-127.0f - 30.0f * x)))
           / (12.0f * x * (1.0f + x) * x2 * x2) + logterm;
  }

 
public:
  PoissonGammaIntegrator(string nodeName, ParameterSet params) :
    BufferedNode(nodeName, params)
  {
    id_input_spec  = addInput("SPEC");
    id_input_sparse  = addInput("SPARSE");
    id_output = addOutput("OUTPUT");

    M = dereference_cast<int>(parameters.get("NUM_CHANNEL"));
    F = dereference_cast<int>(parameters.get("LENGTH")) / 2 + 1;

    scale = dereference_cast<float>(parameters.get("GLOBAL_SCALE"));

    a0 = 0.1, b0.setOnes(F);
    a1 = 1, b1 = 1;
    a2 = 1, b2 = 1;

    s_a.setOnes(F);
    s_b.setOnes(F);

    g_a.setOnes(M);
    g_b = 1;

    e_a.setOnes(F, M);
    e_b = 1;
  }


  void calculate(int output_id, int count, Buffer &out)
  {
    // translate RCPtr to Eigen
//    RCPtr<HARK::Matrix<Complex> > p_input_spec = getInput(id_input_spec, count);
    RCPtr<HARK::Matrix<complex<Real>> > p_input_spec = getInput(id_input_spec, count);
    Eigen::Map<const MatrixXc> input_spec((*p_input_spec)[0], F, M);

    RCPtr<HARK::Matrix<Real> > p_input_sparse = getInput(id_input_sparse, count);
    Eigen::Map<const MatrixXr> y((*p_input_sparse)[0], F, M);

//    RCPtr<HARK::Matrix<Complex> > p_output(new HARK::Matrix<Complex>(1, F));
    RCPtr<HARK::Matrix<complex<Real>> > p_output(new HARK::Matrix<complex<Real>>(1, F));
    Eigen::Map<MatrixXc> output((*p_output)[0], F, 1);

    (*(outputs[id_output].buffer))[count] = p_output;

    cout << scale * y.sum() / (F * M) << endl;

    VectorXr s;
    VectorXr gs, gg;
    MatrixXr ge;
    gs = (s_a.unaryExpr(std::ptr_fun(digamma)) -
	  s_b.array().log().matrix()).array().exp();
    gg = (g_a.unaryExpr(std::ptr_fun(digamma)) -
	  VectorXr::Constant(M, log(g_b))).array().exp();
    for (int it=0;it<50;it++) {
      ge = (e_a.unaryExpr(std::ptr_fun(digamma)) -
	    MatrixXr::Constant(F, M, log(e_b))).array().exp();

      MatrixXr gggs = gg.transpose().replicate(F, 1).cwiseQuotient(gs.replicate(1, M));
      MatrixXr ylmgs = y.cwiseProduct(gggs)
	.cwiseQuotient(gggs + ge + MatrixXr::Constant(F, M, EPS));

      // update s
      s_a = VectorXr::Constant(F, a0) + scale * ylmgs.rowwise().sum();
      s_b = b0 + VectorXr::Constant(F, g_a.sum() / g_b);

      s = s_a.cwiseQuotient(s_b);

      // update g
      g_a = VectorXr::Constant(M, a1) + scale * ylmgs.colwise().sum().transpose();
      g_b = b1 + s.sum();
      
      gs = (s_a.unaryExpr(std::ptr_fun(digamma)) -
	    s_b.array().log().matrix()).array().exp();
      gg = (g_a.unaryExpr(std::ptr_fun(digamma)) -
	    VectorXr::Constant(M, log(g_b))).array().exp();
      gggs = gg.transpose().replicate(F, 1).cwiseQuotient(gs.replicate(1, M));
      MatrixXr ylme = y.cwiseProduct(ge)
	.cwiseQuotient(gggs + ge + MatrixXr::Constant(F, M, EPS));

      // update e
      e_a = MatrixXr::Constant(F, M, a2) + scale * ylme;
      e_b = b2 + 1;

      // update b0
      b0 = VectorXr::Constant(F, a0).cwiseQuotient(s);
    }

    cout << g_a.transpose() / g_b << endl;

    output = (g_a(0) / g_b / scale) * s
      .cwiseProduct(input_spec.col(0))
      .cwiseQuotient(input_spec.col(0).cwiseAbs());
  }
};
